-- Handy shortcuts
local wv = BeneCast.WorkVars;
local BeneCastButton = BeneCast.Button;
local BeneCastPanel = BeneCast.Panel;
local BeneCastPanelDropDown = BeneCast.Panel.DropDown;
local hp = BeneCast.Helper;
local L = BeneCast.Localisation;
local const = BeneCast.Constants;

-- *****************************************************************************
-- Helper functions, non benecast-specific
-- *****************************************************************************

-- Function to clear the table given as the parameter
hp._ClearTable = function(Table, recursively)

	if Table ~= nil and type(Table) == 'table' then
		-- Removes every element of the table starting at the end
		local k, v = next(Table);
		local k2, v2;
		while k ~= nil do
			if type(v) == 'table' and recursively then
				hp._ClearTable(v, true);
				hp.del(v);
			end
			k2, v2 = next(Table,k);
			Table[k] = nil;
			k, v = k2, v2;
		end
		return Table;
	else
		return {};
	end
	
end

hp._CopyTable = function(Table,recursive)
	if Table ~= nil then
		local new = hp.new();
		for k, v in pairs(Table) do
			if (type(v) == 'table') and recursive then
				new[k] = hp._CopyTable(v,true);
			else
				new[k] = v;
			end
		end
		return new;
	else
		return nil;
	end
end

hp._CopyTableToExisting = function(source,destination,recursive)
	if source == nil or destination == nil then
		return;
	end
	for k, v in pairs(source) do
		if (type(v) == 'table') and recursive then
			destination[k] = hp._CopyTable(v,true);
		else
			destination[k] = v;
		end
	end
	return destination;
end

-- Function to determine if a unit is in your raid group
hp._UnitInRaid = function(unit)

	return UnitPlayerOrPetInRaid(unit);
	
end

-- Function to determine if a unitid is 'raid'..x
hp._UnitIsRaidID = function(unit)

	if string.find(unit,'raid') and not string.find(unit,'raidpet') then
		return true;
	end
	return nil;
	
end

-- Function to determine if a unitid is x..'pet'..y
hp._UnitIsPetID = function(unit)

	return string.find(unit, 'pet');
	
end

-- Function to determine if a unitid is 'raidpet'..x
hp._UnitIsRaidPetID = function(unit)

	return string.find(unit, 'raidpet');
	
end

-- Function to determine if a unit is a party pet
hp._UnitIsPartyPet = function(unit)

	return not UnitIsPlayer(unit) and UnitPlayerOrPetInParty(unit);
	
end

-- Function to determine if a unit is a raid pet
hp._UnitIsRaidPet = function(unit)

	return not UnitIsPlayer(unit) and UnitPlayerOrPetInRaid(unit);
	
end

-- Function to determine the owner of a raid pet
hp._GetRaidPetOwnerID = function(unit)
	return string.gsub(unit,'pet','');
end

-- Function to set a lot of attributes in one time
hp._SetManyAttributes = function(self, ...)
	for i=1,select("#", ...),2 do
		local att,val = select(i, ...);
		if not att then
			return;
		end
		self:SetAttribute(att,val);
	end
end

-- Function to get the key of the table member who's value = val
hp._GetKeyForVal = function(Table,val)
	for k, v in pairs(Table) do
		if v == val then
			return k;
		end
	end
end


-- Function to get the key of the table member that has a certain table member with value = val
hp._GetKeyForValInMember = function(Table,member,val)
	for k, v in pairs(Table) do
		if type(v) == 'table' and v[member] == val then
			return k;
		end
	end
end

-- Function to set the value of the table member who's value = oldval
-- 'count' determines how many set's may be done, nil = infinite
hp._ReplaceValues = function(Table,oldval,newval,count)
	local k, v = next(Table);
	while k ~= nil do
		if v == oldval then
			Table[k] = newval;
			if count ~= nil then
				count = count - 1;
				if count <= 0 then
					do break end;
				end
			end
		end
		k, v = next(Table, k);
	end
end

-- Function to set the value of a table member's member that has another/the same member with value = oldval
-- 'count' determines how many set's may be done, nil = infinite
hp._ReplaceValuesByMember = function(Table,oldmember,oldval,newmember,newval,count)
	local k, v = next(Table);
	while k ~= nil do
		if type(v) == 'table' and v[oldmember] == oldval then
			Table[k][newmember] = newval;
			if count ~= nil then
				count = count - 1;
				if count <= 0 then
					do break end;
				end
			end
		end
		k, v = next(Table, k);
	end
end

-- Recycling functions
-- Swiped from AceGUIWidget-TreeGroup
do
	local pool = setmetatable({},{__mode='k'})
	hp.new = function()
		local t = next(pool)
		if t then
			pool[t] = nil
			return t
		else
			return {}
		end
	end
	hp.del = function(t)
		for k in pairs(t) do
			t[k] = nil
		end	
		pool[t] = true
	end
end

-- Version of table.getn that gets a count of ALL members
hp._CountMembers = function(Table)

	local cnt = 0;
	local k, v = next(Table);
	while k ~= nil do
		-- v is probably never nil in this case, but just to be sure...
		if v ~= nil then
			cnt = cnt + 1;
		end
		k, v = next(Table,k);
	end
	return cnt;

end

-- Like the isnull function in SQL, returns a substitute value if first value is nil
hp._isnil = function(value, repl)
	if value == nil then
		return repl;
	else
		return value;
	end
end