-- Handy shortcuts
local wv = BeneCast.WorkVars;
local BeneCastButton = BeneCast.Button;
local BeneCastPanel = BeneCast.Panel;
local BeneCastPanelDropDown = BeneCast.Panel.DropDown;
local hp = BeneCast.Helper;
local L = BeneCast.Localisation;
local const = BeneCast.Constants;

-- Function to set orientation of buttons
function BeneCastPanel:OrientButtons()
	
	if self.unit == nil then
		BeneCast:Debug('BeneCast.Panel:OrientButtons(): unit is nil');
		return;
	end

	if self.pf then
		BeneCast:Debug('BeneCast.Panel:OrientButtons(): pf is nil');
		return;
	end
	
	local profile, panelconf
	profile, panelconf = BeneCast:GetProfile(self.pf, self.unit);
	if profile == nil or panelconf == nil then
		return;
	end
	
	if not profile.orientation then
		profile.orientation = 'RIGHT';
	end
	
	if not panelconf.orientation then
		panelconf.orientation = profile.orientation;
	end

	if InCombatLockdown() then
		return;
	end

	local button = self.next;
	local previousbutton = self;
	while button ~= nil do
		
		button:ClearAllPoints();
		if panelconf.orientation == 'RIGHT' then
			button:SetPoint('LEFT',previousbutton,'RIGHT',3,0);
		elseif panelconf.orientation == 'DOWN' then
			button:SetPoint('TOP',previousbutton,'BOTTOM',0,-3);
		elseif panelconf.orientation == 'UP' then
			button:SetPoint('BOTTOM',previousbutton,'TOP',0,3);
		elseif panelconf.orientation == 'LEFT' then
			button:SetPoint('RIGHT',previousbutton,'LEFT',-3,0);
		end
		
		previousbutton = button;
		button = button.next;
		
	end
	
	self:ResizeButtons();

end

-- Function to resize the buttons of one panel
function BeneCastPanel:ResizeButtons()

	if InCombatLockdown() then
		return;
	end
	
	if self.unit == nil then
		BeneCast:Debug('BeneCast.Panel:OrientButtons(): unit is nil');
		return;
	end
	
	if self.pf then
		BeneCast:Debug('BeneCast.Panel:OrientButtons(): pf is nil');
		return;
	end
	
	local profile = BeneCast:GetProfile(self.pf, self.unit);
	if profile == nil then
		return;
	end
	
	local buttonsize = profile.buttonsize;
	local button = self.next;

	-- Resize the buttons based on the saved size for every button in a panel
	while button ~= nil do
	  button:Resize(buttonsize);
	  button = button.next;
	end

end

function BeneCastPanel:OnHide()

	if ( self.isMoving ) then
		self:StopMovingOrSizing();
		self.isMoving = false;
	end
	
end

function BeneCastPanel:OnLoad()

	-- set our default colors
	self:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	self:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	self.locked = 1;

end

function BeneCastPanel:OnMouseDown(arg1)

	local profile = BeneCast:GetProfile(self.pf);
	if ( ( ( not self.locked ) or ( self.locked == 0 ) or profile.unlockbuttons ) and ( arg1 == 'LeftButton' ) ) then
		self:StartMoving();
		self.isMoving = true;
	end
	
end

function BeneCastPanel:OnMouseUp(arg1)

	if arg1 == "RightButton" then
		ToggleDropDownMenu(1, nil, self.DropDown, self:GetName(), 0, 0);
	end
	if arg1 == "LeftButton" then
		local profile, panelpf = BeneCast:GetProfile(self.pf, self.unit);
		if ( self.isMoving ) then
			self:StopMovingOrSizing();
			self.isMoving = false;
			if panelconf then
				panelpf.point = 'TOPLEFT';
				panelpf.relativepoint = 'TOPLEFT';
				local parent = self:GetParent();
				-- The values stored are unscaled values (real pixels)
				-- When being applied they will need to be adjusted to the scale between the frame and his parent
				panelpf.x = ( self:GetLeft() * self:GetEffectiveScale() ) - 
							( parent:GetLeft() * parent:GetEffectiveScale() ) + 0.5;
				panelpf.x = math.floor(panelpf.x);
				panelpf.y = ( self:GetTop() * self:GetEffectiveScale() ) -
							( parent:GetTop() * parent:GetEffectiveScale() ) + 0.5;
				panelpf.y = math.floor(panelpf.y);
				-- Move the panel to where we saved it, this so any errors in above code or the placement code
				--  would be immediately apparent instead of on next login
				BeneCast.SnapTo:SnapPanel(parent, self, panelpf);
			else
				self:Hide();
			end
		end
	end
	
end

function BeneCastPanel:OnEnter()
	if wv.Setup.debug then
		local unitname, framename;
		
		framename = this:GetName();
		unitname = UnitName(self.unit);
		if not unitname then
			unitname = 'nil';
		end
		BeneCast:Debug('Mousing over ' .. framename .. ', unit = ' .. self.unit .. ', name is ' .. unitname);
	end
end

-- Function to create a new Panel based on the xml-template
-- Panel is stored in BeneCast.WorkVars.Panels[profile][unit]
function BeneCastPanel:NewPanel(pfid, unit)

	local panelname = 'BeneCastPanel_' .. pfid .. '_' .. unit;
	local panel = CreateFrame('Frame', panelname, UIParent, 'BeneCastPanelTemplate');
	hp._CopyTableToExisting(self,panel,true);
	panel:SetScript('OnHide',panel.OnHide);
	panel:SetScript('OnLoad',panel.OnLoad);
	panel:SetScript('OnMouseDown',panel.OnMouseDown);
	panel:SetScript('OnMouseUp',panel.OnMouseUp);
	panel:SetScript('OnEnter',panel.OnEnter);
	panel.DropDown = getglobal(panelname .. 'DropDown');
	hp._CopyTableToExisting(self.DropDown,panel.DropDown,true);
	panel.DropDown.panel = panel;
	panel.DropDown:SetScript('OnLoad',panel.DropDown.OnLoad);
	panel.DropDown:SetScript('OnShow',panel.DropDown.OnShow);
	panel.pf = pfid;
	panel.unit = unit;
	panel:SetAttribute('unit',unit);
	panel:Hide();
	panel:SetFrameStrata('LOW');
	if wv.Panels[pfid] == nil then
		wv.Panels[pfid] = hp.new();
	end
	wv.Panels[pfid][unit] = panel;
	return panel;

end

function BeneCastPanel:ScanButtons()
	
	hp._ClearTable(self.buttons, false);
	local button = self.next;
	local seqno = 1;
	while button do
		tinsert(self.buttons, button);
		button.seqno = seqno;
		seqno = seqno + 1;
		button = button.next;
	end
	
end

-- Function used to update buttons, works for all possible units
function BeneCastPanel:Update()

	if self.unit == nil then
		BeneCast:Debug('BeneCastPanel:Update(): unit is nil');
		return;
	end
	
	if self.pf == nil then
		BeneCast:Debug('Panel for unit ' .. unit .. ' has no profile');
		if self:IsShown() and not InCombatLockDown() then
			self:Hide();
		end
		return;
	end
	
	if self.pf ~= BeneCast:GetCurrentProfileID() then
		if self:IsShown() and not InCombatLockDown() then
			self:Hide();
		end
		return;
	end
	
	local profile = BeneCast:GetProfile(self.pf);
	
	if not wv.AttachedTargets[self.unit] then
		if unit == 'player' or unit == 'target' or unit == 'targettarget' then
			BeneCast:Debug('Not updating '..self.unit..'-panel because it was not previously attached.');
		end
		return;
	end
	wv.AttachedTargets[self.unit] = false;
	
	if not InCombatLockdown() then
		self:Show();
	end
	
	-- Unit statusses
	local disableall = false;
	local unitexists = true;
	local unitisfriend = true;
	local hidepanel = false;

	if ( unit == 'targettarget' and not profile.showtargettarget ) then
		disableall = true;
		hidepanel = true;
	end
	
	-- if this member is non existant the parent frame will hide the panel for us
	if not UnitExists(self.unit) then
		unitexists = false;
		disableall = true;
	end
		
	local name = UnitName(self.unit);
	
	-- Hide this BeneCastPanel frame and return if this member is a pet and ShowPets is not enabled
	-- Except if the player is directly targetting the pet
	if ( self.unit ~= 'target' and self.unit ~= 'targettarget' ) and unitexists and hp._UnitIsPetID(self.unit) then
		if not profile.showpets then
			disableall = true;
			hidepanel = true;
		end
	end
	
	-- This is a hide due to settings, hide and exit if possible
	if hidepanel then
		if not InCombatLockdown() then
			self:Hide();
			return;
		end
	end
	
	-- Remember if member is mindcontrolled
	local curemagic = false;
	if unitexists then
		if UnitIsCharmed(self.unit) then
			curemagic = true;
		end
	end
	
	-- Disable this BeneCastPanel frame if member is not friendly
	if unitexists and (self.unit == 'target' or self.unit == 'targettarget') then
		if ( not UnitIsFriend('player', self.unit) and not curemagic ) then
			disableall = true;
			unitisfriend = false;
		end
	end
	
	-- If the target is invalid due to a temporary game-status keep it hidden if not in combat
	if not unitisfriend or not unitexists or disableall then
		if not InCombatLockdown() and not wv.ConsiderInCombat then	
			self:Hide();
			return;
		end
	end
	
	if unitexists then
		if ( UnitIsGhost('player') or not UnitIsConnected(self.unit) ) then
			disableall = true;
		end
	end
	
	-- If the player can ressurect this member then show the res button and return
	-- 	If there is a buff showing (UnitBuff()) then the player _cannot_ be dead, or they
	-- 	are a priest in Spirit of Redemption form, and so they shouldn't be ressurected.
	-- 	DEAD = No buffs.  Hunters do not need a special case here.
	-- 	We don't bother with ghosts, because we can't cast it on their name, we have to find
	-- 	their body.
	local showres = false;
	if unitexists and unitisfriend then
		if ( UnitIsDead(self.unit) and not UnitBuff(self.unit, 1) ) then
			showres = true;
		end
	end
	
	-- Otherwise show the buttons. Show the BeneCastPanel frame first
	if not InCombatLockdown() then
		self:Show();
	end
	
	local effectsfound = hp.new();
	local overrides = hp.new();
	local enables = hp.new();
	
	-- Check to see if the unit can be cured of a harmful effect
	-- Iterator for debuffs
	local i = 1;
	
	-- Variables to determine if cures should be shown
	-- This is done in case there are multiples of one kind of effect
	local poisonshow = nil;
	local diseaseshow = nil;
	local magicshow = nil;
	local curseshow = nil;
	
	if unitexists and unitisfriend then
		local debuffname, dispeltype, duration;	
		debuffname, _, _, _, dispeltype, duration = UnitDebuff(self.unit, i);
		while debuffname do
			
			local spellinfo = wv.SpellInfo[debuffname];
			if spellinfo then
				effectsfound[spellinfo.group] = {
					type = 'debuff',
					duration = duration,
				};
				if spellinfo.overrides then
					for k, name in pairs(spellinfo.overrides) do
						overrides[name] = {
							type = 'override',
							duration = duration,
						};
					end
				end
				if spellinfo.enables then
					for k, name in pairs(spellinfo.enables) do
						enables[name] = true;
					end
				end
			end
			if dispeltype == 'Poison' then
				poisonshow = {
					type = 'debufftype',
					duration = duration,
				};
			elseif ( dispeltype == 'Disease' ) then
				diseaseshow = {
					type = 'debufftype',
					duration = duration,
				};
			elseif ( dispeltype == 'Magic' ) then
				magicshow = {
					type = 'debufftype',
					duration = duration,
				};
			elseif ( dispeltype == 'Curse' ) then
				curseshow = {
					type = 'debufftype',
					duration = duration,
				};
			end
			i = i + 1;
			debuffname, _, _, _, dispeltype, duration = UnitDebuff(self.unit, i);
		end
	end
	
	-- Check to see if the unit lacks castable buffs
	
	-- Update the current stance, since not all stances are detectable through buffs (mostly only for warriors though)
	if self.unit == 'player' then
		local stancecode = 'none';
		i = 1;
		local name, active, spellinfo;
		_, name, active = GetShapeshiftFormInfo(i);
		while name do
			if active then
				spellinfo = wv.SpellInfo[name];
				if spellinfo then
					if spellinfo.type == 'stance' then
						stancecode = spellinfo.code;
					end
				end
				do break end;
			end
			i = i + 1;
			_, name, active = GetShapeshiftFormInfo(i);
		end
		wv.CurrentStance = stancecode;
	end
	
	-- Reset the iterator for buffs
	i = 1;
	if unitexists and unitisfriend then
	
		-- Cycle through the Units Buffs to find if the effect is on the unit
		local buffname, duration, isMine;
		buffname, _, _, _, _, duration, _, isMine = UnitBuff(self.unit, i);
		while buffname do
		
			local spellinfo = wv.SpellInfo[buffname];
			if spellinfo and spellinfo.type ~= 'stance' then
				effectsfound[spellinfo.group] = {
					type = 'buff',
					duration = duration,
				};
				if spellinfo.overrides then
					for k, name in pairs(spellinfo.overrides) do
						overrides[name] = {
							type = 'override',
							duration = duration,
						};
					end
				end
				if spellinfo.enables then
					for k, name in pairs(spellinfo.enables) do
						enables[name] = true;
					end
				end
				if spellinfo.uninteractible and self.unit ~= 'player' then
					-- unit is uninteractible, disable all buttons
					disableall = true;
				end
			end
			
			i = i + 1;
			buffname, _, _, _, _, duration, _, isMine = UnitBuff(self.unit, i);
		end
	end
	
	-- Get weapon enchants
	if ( member == 'player') then
		if MHWeaponEnchant ~= nil then
			effectsfound[MHWeaponEnchant.group] = {
				type = 'enchant',
				duration = MHWeaponEnchant.duration,
			};
		end
		if OHWeaponEnchant ~= nil then
			effectsfound[OHWeaponEnchant.group] = {
				type = 'enchant',
				duration = OHWeaponEnchant.duration,
			};
		end
	end
	
	-- Overrides make it look like a buff is in effect, disabling the spell
	-- Mostly used for groupbuffs to disable the non-group variant
	for name, v in pairs(overrides) do
		local spellinfo = wv.SpellInfo[name];
		if spellinfo then
			if not effectsfound[spellinfo.group] then
				effectsfound[spellinfo.group] = {
					type = v.type,
					duration = v.duration,
				};
			end
		end
	end
	
	-- Enables make it look like a buff is not in effect, enabling the spell, the spell is usually a consume or a cure
	-- Like Swiftmend lighting up when Rejuvenation/Regrowth is up
	for name, yes in pairs(enables) do
		local spellinfo = wv.SpellInfo[name];
		if spellinfo then
			if effectsfound[spellinfo.group] then
				effectsfound[spellinfo.group].type = 'enable';
			end
		end
	end
	
	local showallbuttons;
	if InCombatLockdown() or wv.ConsiderInCombat then
		showallbuttons = true
	else
		showallbuttons = profile.showallbuttons;
	end
	
	-- Get the config per unit type
	local config, configname;
	if self.unit == 'player' then
		configname = 'player';
	elseif self.unit == 'pet' then
		configname = 'pet';
	elseif hp._UnitIsPartyPet(self.unit) then
		configname = 'partypet';
	elseif UnitPlayerOrPetInParty(self.unit) then
		configname = 'party';
	else
		configname = 'other';
	end
	config = profile.config[configname];
	
	-- Build the buttons from the config
	local buttons = hp.new();
	
	for groupname, setting2 in pairs(config) do
		local setting = hp._CopyTable(setting, true);
		if setting.enabled then
			local action, show, flash, duration, effectfound;
			-- Find the action in the known spells
			action = hp._CopyTable(wv.Spells[groupname],true);
			if action == nil then
				-- TODO: check in known items
			end
			if action == nil then
				-- TODO: check in macro's
			end
			if action == nil then
				if groupname == 'assist' then
					action = hp._CopyTable(BeneCast.Assist,true);
				end
				if groupname == 'focus' then
					action = hp._CopyTable(BeneCast.Focus,true);
				end
			end
			-- Double-check if the action can be executed on the target
			if action and action.target and const.TargetTypes[action.target][configname] then
				-- the code first assumes that a button will be completely hidden (not faded) if it's effect is already active
				-- then at the end, if showallbuttons is true, a button that has .show = false is faded instead
				effectfound = effectsfound[groupname] and (effectsfound[groupname].type ~= 'enable' and effectsfound[groupname].type ~= nil);
				flash = effectfound and action.flash;
				if effectsfound[groupname] then
					duration = effectsfound[groupname].duration;
				end
				if (action.type == 'rez') then
					if (not wv.ConsiderInCombat and showrez) then
						show = true;
					end
				elseif (action.type == 'brez') then
					if (wv.ConsiderInCombat and showrez) then
						show = true;
					end
				elseif action.type == 'cure' then
					for k, curetype in pairs(action.cures) do
						if curetype == 'curse' and curseshow then
							show = true;
						elseif curetype == 'disease' and diseaseshow then
							show = true;
						elseif curetype == 'poison' and poisonshow then
							show = true;
						elseif curetype == 'magic' and magicshow then
							show = true;
						end
					end
				elseif action.type == 'consume' then --consume's have to be specifically enabled by the effect they consume
					show = enables[action.group];
				else
					show = not effectfound;
				end
				if show or showallbuttons then
					if profile.flashasfade then
						action.show = (show and not flash);
						action.flash = nil;
					else
						action.show = show;
						action.flash = flash;
					end
					action.duration = duration;
					action.setting = hp._CopyTable(setting);
					table.insert(buttons,action);
				end
			end
		end
	end
	
	-- Sort the buttons based on their type, sorting property and name
	table.sort(buttons,function(a,b)
		local atypekey, btypekey;
		if a.type then
			atypekey = hp._GetKeyForVal(const.TypesSorted,a.type);
		end
		if b.type then
			btypekey = hp._GetKeyForVal(const.TypesSorted,b.type);
		end
		-- Odd ones go in the back
		if atypekey == nil then
			atypekey = 1000;
		end
		if btypekey == nil then
			btypekey = 1000;
		end
		if atypekey < btypekey then
			return true;
		elseif atypekey == btypekey then
			if a.sorting < b.sorting then
				return true;
			elseif a.sorting == b.sorting then
				return (a.name < b.name);
			end
		end
		return false;
	end);
	
	-- Change the panel

	-- Clear the buttons that flash for this panel
	if not mv.ButtonsThatFlash[self.pf] then
		mv.ButtonsThatFlash[self.pf] = hp.new();
	end
	mv.ButtonsThatFlash[self.pf][self.unit] = hp._ClearTable(mv.ButtonsThatFlash[self.pf][self.unit], false);
	
	local button, prevbutton = panel;
	for i, action in pairs(buttons) do
		button = prevbutton.next;
		if not InCombatLockdown() then
			if not button then
				-- This either makes a brand new button or gets one back from retirement
				button = BeneCastButton:NewButton(panel,prevbutton);
				button:Show();
			end
			button:SetAction(action);
		end
		if button then
			button:SetFade(action.show);
			if action.flash then
				BeneCast:AddButtonThatFlashes(button);
			end
		end
		prevbutton = button;
	end
	-- The rest of the buttons are retired
	if not InCombatLockdown() then
		button = prevbutton.next;
		while button do
			nextbutton = button.next;
			button:Retire();
			button = nextbutton;
		end
	end
	
	-- There is no reason that we would hide this later, so we can check this panel each
	-- 	time we have a reason to suspect a change.
	wv.AttachedTargets[self.unit] = true;
	
	-- Resize the BeneCastPanelFrame based on the number of buttons
	self:OrientButtons();
	
	hp.del(effectsfound);
	hp.del(overrides);
	hp.del(enables);
	hp.del(buttons);
		
end

function BeneCastPanel:SetLock() 

	self.locked = not self.locked;
	
end

-- Function to rotate a panel clockwise
function BeneCastPanel:SetOrientation()

	BeneCast:Debug('BeneCastPanel:SetOrientation() called');
	local pf, panelpf = BeneCast:GetProfile(self.pf, self.unit);
	if pf == nil or panelpf == nil then
		self:Hide();
	else
		if panelpf.orientation == nil then
			panelpf.orientation = 'RIGHT';
		end
		-- The table contains the next value in the rotation
		panelpf.orientation = const.Orientations[panelpf.orientation].next;
		
		BeneCast:Debug('New value wv.Profiles[' .. self.pf .. '][' .. self.unit .. '].orientation = ' ..
			panelpf.orientation);
		self:OrientButtons();
	end

end

-- *****************************************************************************
-- BeneCastPanelMenu functions
-- *****************************************************************************

function BeneCastPanelDropDown:Initialize()
	
	local pf, panelpf = BeneCast:GetProfile(self.panel.pf, self.panel.unit);
	if pf == nil or panelpf == nil then
		return;
	end
	local orientation = panelpf.orientation;
	if orientation == nil then
		orientation = L['Undefined'];
	else
		orientation = const.Orientations[orientation].translation;
	end
	local info = hp.new();
	
	-- Add Title Button
	if ( UnitExists(self.panel.unit) ) then
		info.text = UnitName(self.panel.unit);
	else
		info.text = L['Unknown'];
	end
	info.isTitle = 1;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	-- Add Lock Button
	hp._ClearTable(info, false);
	info.text = L['Locked'];
	info.value = self.panel;
	info.checked = self.panel.locked;
	info.func = self.panel.SetLock;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	-- Add Orientation Button
	hp._ClearTable(info, false);
	info.text = L['Change Orientation (%1)'](orientation);
	info.value = self.panel;
	info.func = self.panel.SetOrientation;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	hp.del(info);

end

function BeneCastPanelDropDown:OnLoad()

	UIDropDownMenu_Initialize(self, self.Initialize, 'MENU');

end

function BeneCastPanelDropDown:OnShow()

	UIDropDownMenu_Initialize(self, self.Initialize, 'MENU');

end